classdef CAheart < handle
    
    properties (Access = private)
        lin2grid
        grid2lin
    end
    
    properties
        % Nb of rows
        n
        
        % m-by-2 array.
        % kth row contains the indices in the grid of a self-exciting cell
        % e.g. pacemakers_indices = [1,1; 1,2];
        pacemakers_indices
        
        % Cross-membrane voltage
        % n^2-by-1 vector
        V
        % Local timers, n^2-by-1 vector
        timers
        % dV/dt used for trough detection
        Vdot
        
        % Cell model used in this heart
        myocyteModel
        
        % Horizontal resistances. Rh(i,j) is the horizontal resistance for
        % contributing to cell (i,j)
        myocytes
        
    end
    
    %% Private methods
    methods (Access = private)
        
        %% Gradient
        function dx = gradient(obj,allfs, t,x)
            % t: time
            % x: state vector = [n^2 voltages, n^2 local timers]
            nn = obj.n;
            dx = zeros(obj.myocyteModel.dimX*nn^2,1);
            for k=1:nn^2
                [i,j] = ind2sub([nn,nn],k);
                % dvdt(1) is gradient V', dvdt(2) is derivative of local
                % timer = 1
                dvdt = allfs{k}(t,x,...
                    obj.myocytes{i,j},k);
                for d=1:obj.myocyteModel.dimX
                    dx(k+(d-1)*nn^2) = dvdt(d);
                end
            end
        end
        
        %% Event function
        function [val,isterm,direction] = guards(obj,allguards,allevents,t,x)
            % Guard condition evaluation.
            % OUTPUTS:
            %   val(i) -> the value of the ith guard function
            %            1 if the guard is not activated
            %           -1 if the guard is activated
            %   isterm(i) -> 1 if computation stops at guard, 0 otherwise
            %   direction(i) -> 0 if all crossings are 2b computed, 1 if only
            %   increasing crossings count, -1 if only decreasing crossings
            %   count.
            % t: time
            % x: state vector = [n^2 voltages, n^2 local timers]
            nn = obj.n;
            % For guards
            val         = [];
            isterm      = [];
            direction   = [];
            % For (non-terminating) events
            valE = [];
            istermE = [];
            directionE = [];
            for k=1:nn^2
                % Each myocyte has its own guard function (guard
                % conditions). Note that in some modes, the myocyte has
                % more than one guard enabled, so its returned val, isterm
                % and direction will be vectors, one entry per guard.
                [i,j] = ind2sub([nn,nn], k);
                [value, isterminal, Dir] = allguards{k}(t,x,...
                    obj.myocytes{i,j},k);
                val         = [val value];
                isterm      = [isterm isterminal];
                direction   = [direction Dir];
                % non-terminating events
                [valueE, isterminalE, DirE] = allevents{k}(t,x,...
                    obj.myocytes{i,j},k);
                valE         = [valE valueE];
                istermE      = [istermE isterminalE];
                directionE   = [directionE DirE];
            end
            val = [val valE]';
            isterm = [isterm istermE]';
            direction = [direction directionE]';
        end
        
        function [new_modes, xnew] = jump(obj,IE, TE, x, nb_enabled_guards, trough_detection_package)
            % new_modes = CAmodel.jump(IE, TE, x, nb_enabled_guards)
            % Find the new modes for myocytes that jump. IE and TE are the
            % vectors returned by ode45 when supplied with an events
            % function.
            % For us, we have guards which are terminating events, and we
            % have non-terminating events which we simply call 'events'.
            % Both are implemented using ode45's Events capability, and IE
            % contains the indices of the guards and the (non-terminating)
            % events that occurred. TE contains their timestamps.
            %
            % IE
            %   vector of ODE solver events containing indices for guards
            %   and non-terminating events that were satisfied.
            % TE
            %   corresponding vector of times. See ode45 help.
            % x
            %   is the state that satisfied the guards.
            % trough_detection_package
            %   a struct with the following fields, used for trough
            %   detection
            %   tlast
            %       time of last point on the current continuous trajectory
            %       piece which is ending with this jump.
            %   xfirst
            %     is the initial state for the current continuous trajectory piece
            %     (so x is the last point on this trajectory
            %     and xfirst is the first pnt)
            %   tfirst
            %       time of first point xfirst
            % nb_enabled_guards
            %   is the nb of guards enabled in the grid.
            % The val vector returned by guards() thus has length
            % nb_enabled_guards + nb_events, where the latter is currently
            % hard-coded to n^2. IE indexes into val, so if IE(m) >
            % nb_enabled_guards for some m, this means that IE(m) is an
            % event and not a guard.
            %
            % The question is which myocyte the events belong to. This is made
            % complicated by the fact that the nb of event fnts of a given
            % myocyte depends on the mode it's in. E.g. one event (guard)
            % in mode QUIESCENT and two events in mode RRP.
            %
            % new_modes is m-by-2: new_modes(i,1) is a myocyte's linear
            % index and new_modes(i,2) is the new mode of that myocyte.
            % Only myocytes that actually jumped (regardless of destination
            % location) are in here.
            % xnew is the post-reset state.
            nn =obj.n;
            
            IEguards = IE(IE <= nb_enabled_guards);
            TEguards = TE(IE <= nb_enabled_guards);
            % Normally, IE should contain either one event, or several
            % events occuring at the same time, since all myocyte events
            % are declared to be terminal. But I've observed otherwise,
            % which seems to happen when matlab starts a simulation where a
            % previous simulation left off. In that case, the initial point
            % of the second simulation is the point that caused the event
            % that ended the first simulation, and it gets re-detected.
            % Matlab deals with this by ignoring this initial event, but
            % still reports just in case.
            % See also http://stackoverflow.com/questions/7621664/ode-solver-event-location-index-in-matlab
            % This code will ignore the initial event.
            if length(IEguards) >= 2 && any(TEguards(1:end-1)-TEguards(2:end))
                IEguards = IEguards(2:end,:);
                TEguards = TEguards(2:end,:);
                % check again for sanity
                if length(IEguards) >= 2 && any(TEguards(1:end-1)-TEguards(2:end))
                    warning('Come on!!')
                    keyboard
                end
            end
            new_modes   = zeros(length(IEguards),2);
            xnew        = x;
            nmix=1;
            eix=1;
            for j=1:nn
                for i=1:nn
                    k = i+(j-1)*nn; % linear index
                    myo = obj.myocytes{i,j};
                    % First, handle some auxiliary stuff related to trough
                    % detection and not really jump dynamics
                    if obj.myocyteModel.trough_detection_mode(myo.mode)
                        [~, ixminv] = min([obj.myocytes{i,j}.Vtrough, trough_detection_package.xfirst(k), x(k)]);
                        if ixminv == 2
                            obj.myocytes{i,j}.Vtrough = trough_detection_package.xfirst(k);
                            obj.myocytes{i,j}.ttrough = trough_detection_package.tfirst;
                            obj.myocytes{i,j}.troughWasset = 1;
                        elseif ixminv == 3
                            obj.myocytes{i,j}.Vtrough = x(k);
                            obj.myocytes{i,j}.ttrough = trough_detection_package.tlast;
                            obj.myocytes{i,j}.troughWasset = 1;
                        end
                    end
                    nb_guards = obj.myocyteModel.nb_guards(myo.mode);
                    assert(nb_guards > 0)
                    % See if an event of this myocyte occurred.
                    % The range eix,eix+nb_guards-1 gives the range of
                    % indices for the guard fnts of myocyte (i,j)
                    ff=find(eix <= IEguards & IEguards <= eix+nb_guards-1);
                    if ~isempty(ff)>0
                        assert(length(ff)<=1, ...
                            sprintf('Mode %s of myocyte %s (i,j = %d,%d) has more than one satisfied guard, non-deterministic', ...
                            char(myo.mode), obj.myocyteModel.name, i,j))
                        % IEguards indexes into the val vector of entire system, must
                        % make it index into val vector of just this myocyte
                        activatedGuard  = IEguards(ff) - (eix-1);
                        activationTime  = TEguards(ff);
                        assert(activatedGuard > 0)
                        [nm, xr, myo]       = ...
                            obj.myocyteModel.jump(activatedGuard, activationTime, ...
                            myo, x([k,k+nn^2,k+2*nn^2]),k ...
                            );
                        new_modes(nmix,:)   = [k, nm];
                        nmix                = nmix+1;
                        xnew(k)             = xr(1); % voltage
                        xnew(k+nn^2)        = xr(2); % local timer
                        xnew(k+2*nn^2)      = xr(3); % dV/dt
                        obj.myocytes{i,j}   = myo;
                    end
                    eix=eix+nb_guards;
                end
            end
            % Non-terminating events. Here we know exactly what the
            % event's significance is:
            % 1. The first event indicates the voltage dropped below
            % APD90 in mode RRP, indicating the end of the APD and the
            % beginning of the DI.
            % 2. The 2nd event indicates dV/dt ==0 so local minimum
            % IEdi indexes into the val vector of
            % entire system which has nb_enabled_guards+2*nn^2 entries, so
            % subtract nb_enabled_guards to re-index into 1:n^2
            IEdi    = IE(IE>nb_enabled_guards) - nb_enabled_guards;
            TEdi    = TE(IE>nb_enabled_guards);
            % XXX handle case where upstroke2 happens in RRP so APD90 never
            % happens
            for m=1:length(IEdi)
                % every myocyte has 2 non-terminating events, so if
                % index valEdi = 1 2 3 4 5 6 7 8 for a 2-by-2 grid, where
                % (1,2)
                % are nt event fnts for myocyte 1, (3,4) for myocyte 2, etc.
                % Then if IEdi = [2 3 8] then the even indices 2 and 8
                % indicate the 2nd nt event fnt of myocytes 1 and 4 resp.,
                % while 3 is the 1st nt event fnt of myocyte 2.
                if mod(IEdi(m),2)==0
                    % 2nd event: trough voltage
                    ixMyocyte = IEdi(m)/2;
                    [i,j]=ind2sub([nn,nn],ixMyocyte);
                    obj.myocytes{i,j}.Vtrough = x(ixMyocyte);
                    obj.myocytes{i,j}.ttrough = TEdi(m);
                    obj.myocytes{i,j}.troughWasset = 1;
                else
                    % 1st event, below APD90
                    ixMyocyte = (IEdi(m)+1)/2;
                    [i,j]=ind2sub([nn,nn],ixMyocyte);
                    obj.myocytes{i,j}.endofAPD = TEdi(m);
                    % Note the actual APD, here computed, doesn't play a direct
                    % role in determining the next APD
                    obj.myocytes{i,j}.APD = max(obj.myocytes{i,j}.params(:,:,obj.myocyteModel.ixAPDMIN), TEdi(m) - obj.myocytes{i,j}.upstroke_time);
                    obj.myocytes{i,j}.APDwasset = 1;
                end
                
            end
        end
        
    end
    
    
    
    %% Public methods
    methods
        %% constructor
        function heart = CAheart(n,varargin)
            % heart = CAheart(n)
            % Grid is n-by-n
            % heart = CAheart(n,'myocytemodel')
            % By default, myocyte model is MyocyteVisibleEP. Can provide
            % another model that respects the interface.
            % heart = CAheart(n,'',Rh)
            % Rh is n-by-n matrix of horizontal resistances. If empty, uses
            % random values in [1,2], which is still pretty homogeneous
            % heart = CAheart(n,Rh, Rv)
            % Vertical resistances as well
            % heart = CAheart(n,..., cell_params)
            %  provide all the cell parameters
            heart.n = n;
            
            va = 1;
            if nargin >= va+1 && ~isempty(varargin{va})
                mm = varargin{va};
            else
                mm = 'VisibleEP';
            end
            if strcmp(mm,'VisibleEP')
                heart.myocyteModel = MyocyteVisibleEP.instance(n, 'use restitution');
            else
                error('Unsupported myocyte model')
            end
            
            va = va+1;
            if nargin >= va+1 && ~isempty(varargin{va})
                Rh = varargin{va};
                assert(size(Rh,1)==n, 'Rh must be n-by-n. You provided n=%d, and Rh %d-ny-%d\n',n,size(Rh,1),size(Rh,2));
            else
                Rh = 1+rand(heart.n,heart.n);
            end
            assert(isequal(size(Rh), [heart.n,heart.n]))
            
            va = va+1;
            if nargin >= va+1 && ~isempty(varargin{va})
                Rv = varargin{va};
                assert(size(Rh,1)==n, 'Rv must be n-by-n. You provided n=%d, and Rv %d-ny-%d\n',n,size(Rh,1),size(Rh,2));
            else
                Rv = 1+rand(heart.n,heart.n);
            end
            assert(isequal(size(Rv), [heart.n,heart.n]))
            
            va = va+1;
            if nargin >= va+1 && ~isempty(varargin{va})
                cell_params = varargin{va};
            else
                cell_params = heart.myocyteModel.generate_params(heart.n,0);
            end
            
            va = va+1;
            is_pacemaker = zeros(heart.n,heart.n);
            if nargin >= va+1 && ~isempty(varargin{va})
                heart.pacemakers_indices = varargin{va};
                npmcells = size(heart.pacemakers_indices, 1);
                for kpm=1:npmcells
                    is_pacemaker(heart.pacemakers_indices(kpm,1),heart.pacemakers_indices(kpm,2)) = 1;
                end
            else
                npmcells = ceil(0.1*heart.n)^2;
                heart.pacemakers_indices = zeros(npmcells,2);
                kpm=1;
                for i=1:sqrt(npmcells)
                    for j = 1:sqrt(npmcells)
                        heart.pacemakers_indices(kpm,1) = i;
                        heart.pacemakers_indices(kpm,2) = j;
                        is_pacemaker(i,j)=1;
                        kpm = kpm+1;
                    end
                end
            end
            
            %% Cell Neighborhood structure
            [horiz_cell_nhbrs, verti_cell_nhbrs] = heart.myocyteModel.create_4nhbr(n);
            
            %% Myocyte struct
            mos = cell(heart.n, heart.n);
            for i=1:heart.n
                for j=1:heart.n
                    myo = struct('Rh', Rh(i,j),...
                        'Rv', Rv(i,j),...
                        'is_pacemaker', is_pacemaker(i,j), ...
                        'mode', heart.myocyteModel.default_mode, ...
                        'params', cell_params(i,j,:), ...
                        'horiz_nhbrs', horiz_cell_nhbrs{i,j},...
                        'verti_nhbrs', verti_cell_nhbrs{i,j},...
                        'Vtrough', 0, ...
                        'ttrough', 0, ...
                        'troughWasset', 0, ...
                        'upstroke_time', 0, ...
                        'endofAPD', -100, ... % to ensure first DI is not freakishly short
                        'APD', 0, ...   % APD = endofAPD_n - upstroke_time_n
                        'APDwasset', 1, ... % boolean. start with it 'set' since we indeed set endofAPD to -100
                        'DI', 0 ...     % DI = upstroke_time_{n+1} - endofAPD_n
                        );
                    mos{i,j} = myo;
                end
            end
            heart.myocytes = mos;
            
            heart.V         = heart.myocyteModel.default_voltage*ones(heart.n^2,1);
            heart.timers    = zeros(heart.n^2,1);
            heart.Vdot      = zeros(heart.n^2,1);
        end
        
        function reset(obj)
            obj.V = obj.myocyteModel.default_voltage*ones(obj.n^2,1);
            obj.timers = zeros(obj.n^2,1);
            obj.Vdot = zeros(obj.n^2,1);
            for i=1:obj.n
                for j=1:obj.n
                    obj.myocytes{i,j}.mode = obj.myocyteModel.default_mode;
                    obj.myocytes{i,j}.endofAPD = -100;
                    obj.myocytes{i,j}.APDwasset = 1;
                    obj.myocytes{i,j}.APD = 0;
                    obj.myocytes{i,j}.upstroke_time = 0;
                    obj.myocytes{i,j}.Vtrough = 0;
                    obj.myocytes{i,j}.ttrough = 0;
                    obj.myocytes{i,j}.troughWasset = 0;
                    obj.myocytes{i,j}.DI = 0;
                end
            end
        end
        
        %% Simulate heart
        function sol = simulate(obj,tf, varargin)
            nn = obj.n;
            solver = 'ode45';
            % allfs and allguards will store handles of functions.
            allfs = cell(nn^2,1);
            allguards = cell(nn^2,1);
            allevents = cell(nn^2,1);
                        prevIE=[];prevTE=[];prevYE=[];
            % The state x consists of n^2 voltages and n^2 local timers.
            sol.XT  = [obj.V', obj.timers', obj.Vdot'];
            % Global time
            sol.T   = 0;
            % Location at each point in time
            sol.LT  = zeros(1,nn^2);
            % Sequence of visited locations/modes by each myocyte
            sol.LOC = cell(nn,1);
            for i=1:nn
                for j=1:nn
                    sol.LT(1,i+(j-1)*nn) = obj.myocytes{i,j}.mode;
                    sol.LOC{i+(j-1)*nn}  = obj.myocytes{i,j}.mode;
                end
            end
            while sol.T(end) < tf
                % Sum of the nb of guards enabled in each myocyte in its
                % current mode = nb all enabled guards in the grid. This
                % depends on the modes in which the myocytes are
                nb_enabled_guards = 0;
                % For each cell, pick the right dynamics and guards at this
                % point in time based on the mode
                for k=1:nn^2
                    [i,j] = ind2sub([nn,nn],k);
                    myo = obj.myocytes{i,j};
                    if myo.is_pacemaker
                        allfs{k}     = obj.myocyteModel.selfexc_dynamics(myo.mode);
                    else
                        allfs{k}     = obj.myocyteModel.dynamics(myo.mode);
                    end
                    allguards{k} = obj.myocyteModel.guards(myo.mode);
                    allevents{k} = obj.myocyteModel.ntevents(myo.mode);
                    nb_enabled_guards = nb_enabled_guards + obj.myocyteModel.nb_guards(myo.mode);
                end
                F = @(t,x) obj.gradient(allfs,t,x);
                G = @(t,x) obj.guards(allguards, allevents, t,x);
                ode_opts = odeset;
                ode_opts = odeset(ode_opts, 'Events', G);
                
                [tt,xx,TE,YE,IE] = feval(solver, F,[sol.T(end),tf],sol.XT(end,:),ode_opts);
                % -----------------------------------------------
                if length(IE) >= 2 && any(TE(1:end-1)-TE(2:end))
                    plot(tt,xx(:,IE))
                    legend('1','2','3')
                    hold on
                    for ii=1:length(IE)
                        [a,b]=ind2sub([nn,nn],IE(ii));
                        %fprintf('Mode (%d,%d) = %d\n',a,b,obj.myocytes{a,b}.mode)
                        plot(TE(ii),YE(ii,IE(ii)),'ro')                        
                    end
                    plot(prevTE(end), prevYE(end,prevIE(end)),'k*')
                    disp('woo!')
                    keyboard
                end
                prevIE = IE;prevTE=TE; prevYE=YE;
                % --------------------------------------------
                % Append. Because we have already appended the initial
                % state and mode at the end of the previous iteration, we
                % skip them on this iteration by using 2:end
                sol.T   = [sol.T; tt(2:end)];
                sol.XT  = [sol.XT; xx(2:end,:)];
                sol.LT  = [sol.LT; repmat(sol.LT(end,:), size(xx,1)-1, 1)];
                if ~isempty(IE)
                    % One or more guards, for one of more myocytes, have
                    % been satisfied
                    % Note that events are not terminating so if we get
                    % here without any guards satisfied this means the
                    % simulation time is over.
                    trough_detection_package = struct('tlast', tt(end), 'xfirst', xx(1,:), 'tfirst', tt(1));
                    [new_modes, xnew] = obj.jump(IE, TE, YE(end,:), nb_enabled_guards, trough_detection_package);
                    sol.XT(end+1, :) = xnew;
                    sol.T(end+1) = sol.T(end);
                    sol.LT(end+1, :) = sol.LT(end,:);
                    for nmix=1:size(new_modes,1)
                        sol.LOC{new_modes(nmix,1)} = [sol.LOC{nmix}, new_modes(nmix,2)];
                        sol.LT(end,new_modes(nmix,1)) = new_modes(nmix,2);
                    end
                end
            end
            
            % Simulation can continue from where it left off
            obj.V        = sol.XT(end,1:nn^2)';
            obj.timers   = sol.XT(end,nn^2+1:2*nn^2)';
            obj.Vdot     = sol.XT(end,2*nn^2+1:end)';
        end
        
        function draw(obj, s)
            nn=obj.n;
            if nargin < 2
                for i=1:nn
                    for j=1:nn
                        x = [j-1; j; j; j-1];
                        y = [nn-i; nn-i;  nn-(i-1); nn-(i-1)];
                        c = obj.myocyteModel.colors{obj.myocytes{i,j}.mode};
                        fill(x,y,c); hold on
                    end
                end
            else
                % s is provided
                for t=1:10:length(s.T)
                    for k=1:nn^2
                        [i,j]=ind2sub([nn,nn],k);
                        x = [j-1; j; j; j-1];
                        y = [nn-i; nn-i;  nn-(i-1); nn-(i-1)];
                        c = obj.myocyteModel.colors{s.LT(t,k)};
                        fill(x,y,c); hold on
                    end
                    title(num2str(t))
                end
            end
        end
        
        function [C,IC]=first_entry(obj,s,ix)
            assert(all(ix<= obj.n^2))
            assert(length(ix)==1)
            for k=ix
                lt = s.LT(:,k);
                IC = find(lt(2:end) ~= lt(1:end-1));
                IC = IC+1; % index into LT
                C  = lt(IC);
                % Add the first point
                IC = [1; IC]; %#ok<AGROW>
                C  = [lt(1); C]; %#ok<AGROW>
            end
        end
        function plot(obj,s, ix, onone)
            % h = CAheart(5);
            % s = h.simulate(1000);
            % h.plot(s, [1,3])
            % plot the trajectories [1,3] each on a figure
            % h.plot(s, [1,3], 1)
            % As before, but plot everything on the same graph
            if nargin < 4
                onone=0;
            end
            pix = 1;
            assert(all(ix<= obj.n^2))
            for k=ix
                x = s.XT(:,k);
                lt = s.LT(:,k);
                IC = find(lt(2:end) ~= lt(1:end-1));
                IC = IC+1; % index into LT
                C  = lt(IC);
                % Add the first point
                IC = [1; IC]; %#ok<AGROW>
                C  = [lt(1); C]; %#ok<AGROW>
                %subplot(length(ix),1,pix)
                if ~onone
                    figure;
                end
                plot(s.T, x);
                hold on
                % Annotation
                if ~onone
                    plot(s.T(IC),x(IC),'ro');
                    for i=1:length(IC)
                        text(s.T(IC(i)), x(IC(i)), num2str(C(i)), 'FontSize', 20)
                    end
                    %plot(s.T,s.XT(:,k+2*obj.n^2),'r-')
                    title(sprintf('%d',k))
                end
                pix = pix+1;
                xlabel('t (ms)')
                ylabel('Vm (mV)')
            end
        end
    end
end

%                     if length(IE) >= 2 && any(TE(1:end-1)-TE(2:end))
%                         mean(sol.T(2:end)-sol.T(1:end-1))
%                         mean(TE(2:end)-TE(1:end-1))
%                         plot(tt,xx(:,IE))
%                         legend('1','2')
%                         hold on
%                         for ii=1:length(IE)
%                             [a,b]=ind2sub([nn,nn],IE(ii));
%                             fprintf('Mode (%d,%d) = %d\n',a,b,obj.myocytes{a,b}.mode)
%                             plot(TE(ii),YE(ii,IE(ii)),'ro')
%                         end
%                         disp('woo!')
%                         keyboard
%                     end
%                     prevIE = IE;prevTE=TE; prevYE=YE;