classdef MyocyteVisibleEP
    
    properties (Access = private, Constant)
        QUIESCENT =1;
        UPSTROKE  =2;
        NOTCH     =3;
        PLATEAU   =4;
        ERP       =5;
        RRP       =6;
        UPSTROKE2 =7;    
    end
    
    properties        
        % d (upstroke, phase 0)
        ixUPSTROKE_SLOPE = 1;
        % g (notch, phase 1)
        ixNOTCH_SLOPE = 2;
        % b (phase 3 ERP)
        ixERP_SLOPE = 3;
        % d2 (upstroke 2)
        ixUPSTROKE2_SLOPE = 4;
        % Vth for depolarization
        ixVTH = 5; 
        % Vmax
        ixVMAX = 6;
        % Vplateau (transition from Notch to Plateau)
        ixVPLATEAU = 7;
        % PD (plateau duration)
        ixPLATEAU_DURATION = 8;
        % Vth2 > Vth (remember Vth is negative)
        ixVTH2 = 9;
        % Vmax2 < Vmax
        ixVMAX2 = 10;
        % Vmin (voltage when quiescent)
        ixVMIN = 11;
        % Vrrp (transition from ERP to RRP when V(i,j) <= Vrrp)
        ixVRRP = 12;
        % APDmax
        ixAPDMAX = 13;
        % APDmin
        ixAPDMIN = 14;
        % dmin = min upstroke velocity
        ixUPSTROKE_SLOPE_MIN = 15;
        % dmax= max upstroke velocity
        ixUPSTROKE_SLOPE_MAX = 16;
        
        % nb of rows in grid in which this cell model is used
        n
        
        % nb of state variables of ths myocyte model
        dimX
        
        % Name of cell model
        name
        
        % Map from cell mode to continuous dynamics x' = f(t,x)
        dynamics
        
        % Map from self-exciting cell mode to continuous dynamics x' = f(t,x)
        selfexc_dynamics
        
        % Map from cell mode to guard conditions
        guards
        
        % Map from cell mode to (non-terminating) event functions
        ntevents
        
        % Nb of Modes
        nmodes
        
        % transitions is a map from mode to its possible successor modes in
        % a discrete transition
        transitions
        
        % mode in which cell starts
        default_mode
        
        default_voltage
        
        colors = {'c','r','k','m','y','g','b'};
        
        % Percentage that defines where the APD ends. E.g. APD lasts until 
        % V(i,j) drops below 10% of its max amplitude |Vmax - Vmin|
        APDpercentage = 0.1;
        
        use_restitution = 0;
        
        % Intrinsic current of self-exciting cells in quiescent mode
        Vinq = 1;
        
        % Intrinsic current of self-exciting cells in RRP mode
        Vinrrp = .1;
                
    end
    
    methods (Access = private)
        % constructor
        function obj = MyocyteVisibleEP(n,varargin)
            obj.n = n;
            obj.name = 'VisibleEP';
            obj.default_mode = MyocyteVisibleEP.QUIESCENT;
            obj.default_voltage = -90;
            obj.dimX = 3;
            if nargin>=2 && strcmp(varargin{1}, 'use restitution')
                obj.use_restitution = 1;
            end
            
            obj.dynamics = containers.Map('KeyType', 'double', 'ValueType', 'any');
            obj.dynamics(MyocyteVisibleEP.QUIESCENT)    = @obj.gradientQUIESCENT;
            obj.dynamics(MyocyteVisibleEP.UPSTROKE)     = @obj.gradientUPSTROKE;
            obj.dynamics(MyocyteVisibleEP.NOTCH)        = @obj.gradientNOTCH;
            obj.dynamics(MyocyteVisibleEP.PLATEAU)      = @obj.gradientPLATEAU;
            obj.dynamics(MyocyteVisibleEP.ERP)          = @obj.gradientERP;
            obj.dynamics(MyocyteVisibleEP.RRP)          = @obj.gradientRRP;
            obj.dynamics(MyocyteVisibleEP.UPSTROKE2)    = @obj.gradientUPSTROKE2;
            
            obj.selfexc_dynamics = containers.Map('KeyType', 'double', 'ValueType', 'any');
            obj.selfexc_dynamics(MyocyteVisibleEP.QUIESCENT)    = @obj.gradient_selfexcQUIESCENT;
            obj.selfexc_dynamics(MyocyteVisibleEP.UPSTROKE)     = @obj.gradient_selfexcUPSTROKE;
            obj.selfexc_dynamics(MyocyteVisibleEP.NOTCH)        = @obj.gradientNOTCH;
            obj.selfexc_dynamics(MyocyteVisibleEP.PLATEAU)      = @obj.gradientPLATEAU;
            obj.selfexc_dynamics(MyocyteVisibleEP.ERP)          = @obj.gradientERP;
            obj.selfexc_dynamics(MyocyteVisibleEP.RRP)          = @obj.gradientRRP;
            obj.selfexc_dynamics(MyocyteVisibleEP.UPSTROKE2)    = @obj.gradientUPSTROKE2;
            
            % Guards are implemented as terminating events
            obj.guards = containers.Map('KeyType', 'double', 'ValueType', 'any');
            obj.guards(MyocyteVisibleEP.QUIESCENT)      = @obj.guardQUIESCENT;
            obj.guards(MyocyteVisibleEP.UPSTROKE)       = @obj.guardUPSTROKE;
            obj.guards(MyocyteVisibleEP.NOTCH)          = @obj.guardNOTCH;
            obj.guards(MyocyteVisibleEP.PLATEAU)        = @obj.guardPLATEAU;
            obj.guards(MyocyteVisibleEP.ERP)            = @obj.guardERP;
            obj.guards(MyocyteVisibleEP.RRP)            = @obj.guardRRP;
            obj.guards(MyocyteVisibleEP.UPSTROKE2)      = @obj.guardUPSTROKE2;
            
            % Events that are not terminating
            obj.ntevents = containers.Map('KeyType', 'double', 'ValueType', 'any');
            obj.ntevents(MyocyteVisibleEP.QUIESCENT)      = @obj.noevent;
            obj.ntevents(MyocyteVisibleEP.UPSTROKE)       = @obj.noevent;
            obj.ntevents(MyocyteVisibleEP.NOTCH)          = @obj.noevent;
            obj.ntevents(MyocyteVisibleEP.PLATEAU)        = @obj.noevent;
            obj.ntevents(MyocyteVisibleEP.ERP)            = @obj.noevent;
            obj.ntevents(MyocyteVisibleEP.RRP)            = @obj.nteventsRRP;
            obj.ntevents(MyocyteVisibleEP.UPSTROKE2)      = @obj.noevent;
            
            obj.nmodes = 7;
            
            obj.transitions = containers.Map('KeyType', 'double', 'ValueType', 'any');
            obj.transitions(MyocyteVisibleEP.QUIESCENT) = MyocyteVisibleEP.UPSTROKE;
            obj.transitions(MyocyteVisibleEP.UPSTROKE) =   MyocyteVisibleEP.NOTCH;
            obj.transitions(MyocyteVisibleEP.NOTCH)    =   MyocyteVisibleEP.PLATEAU;
            obj.transitions(MyocyteVisibleEP.PLATEAU)   =  MyocyteVisibleEP.ERP;
            obj.transitions(MyocyteVisibleEP.ERP)        = MyocyteVisibleEP.RRP;
            obj.transitions(MyocyteVisibleEP.RRP)        = [MyocyteVisibleEP.UPSTROKE2, MyocyteVisibleEP.QUIESCENT];
            obj.transitions(MyocyteVisibleEP.UPSTROKE2)  = MyocyteVisibleEP.PLATEAU;
            
        end
        
        function apd = restitution(obj,DI)
            % Restitution curve
            % apd = a + b/(1 + exp(-(DI-c)/d));
            b=400;
            c=1;
            d=50;
            a=-b/(1+exp(c/d));
            apd = a + b/(1 + exp(-(DI-c)/d));
        end
    end
    
    methods(Static)
        % Concrete implementation.  See Singleton superclass.
        function obj = instance(n, varargin)
            %             persistent uniqueInstance
            %             if isempty(uniqueInstance)
            %                 obj = MyocyteVisibleEP(n, varargin{:});
            %                 uniqueInstance = obj;
            %             else
            %                 obj = uniqueInstance;
            %             end
            obj = MyocyteVisibleEP(n, varargin{:});
        end
        
    end
    
    methods
        
        %% Guards
        function [val, isterm, Dir] = guardQUIESCENT(obj,t,x,myo,k)
            % Guard when QUIESCENT: V(i,j) >= Vth
            % x is the voltage vector of heart
            val = x(k) - myo.params(obj.ixVTH);
            isterm = 1;
            Dir = 1;
        end
        
        function [val, isterm, Dir] = guardUPSTROKE(obj,t,x,myo,k)
            % Upstroke to Notch: V(i,j) >= Vmax
            % x is the voltage vector of heart
            val = x(k) - myo.params(obj.ixVMAX);
            isterm = 1;
            Dir = 1;
        end
        
        function [val, isterm, Dir] = guardNOTCH(obj,t,x,myo,k)
            % Notch to plateau V(i,j) <= Vplateau
            % x is the voltage vector of heart
            val =  x(k) - myo.params(obj.ixVPLATEAU);
            isterm = 1;
            Dir = -1;
        end
        
        function [val, isterm, Dir] = guardPLATEAU(obj,t,x,myo,k)
            % Plateau to ERP: t >= PD
            % x is the voltage vector of heart
            val = x(k+obj.n^2) - myo.params(obj.ixPLATEAU_DURATION);
            isterm = 1;
            Dir = 1;
        end
        
        function [val, isterm, Dir] = guardERP(obj,t,x,myo,k)
            % From ERP to RRP: V(i,j) <= Vrrp
            % x is the voltage vector of heart
            val = x(k) - myo.params(obj.ixVRRP);
            isterm = 1;
            Dir = -1;
        end
        
        function [val, isterm, Dir] = guardRRP(obj,t,x,myo,k)
            % Two guards when in RRP:
            % RRP to Upstroke2: V(i,j) >= Vth2
            val(1)      = x(k) - myo.params(obj.ixVTH2);
            isterm(1)   = 1;
            Dir(1)      = 1;
            % RRP To Quiescent: V(i,j) <= Vmin
            val(2)      = x(k) - myo.params(obj.ixVMIN);
            isterm(2)   = 1;
            Dir(2)      = -1;
        end
        
        function [val, isterm, Dir] = nteventsRRP(obj,t,x,myo,k)
            % Keep track of when the voltage drops below the APD90 value
            % (or whatever is the percentage set by the APDpercentage
            % property)
            % x  is the system's state vector with dimX*nn^2 entries
            % k is linear index of the grid myocyte
            % V(i,j) <= Vmin + perc*(Vmax-Vmin)
            val(1)      = x(k) - (...
                myo.params(obj.ixVMIN) + obj.APDpercentage*(myo.params(obj.ixVMAX) - myo.params(obj.ixVMIN) )...
                );
            isterm(1)   = 0;
            Dir(1)      = -1;
            % Minimum (trough voltage) before depolarization: V'==0
            val(2)      = x(k+2*obj.n^2);
            isterm(2)   = 0;
            Dir(2)      = 1;            
        end
        
        function [val, isterm, Dir] = guardUPSTROKE2(obj,t,x,myo,k)
            % Upstroke 2 to Plateau V(i,j) >= Vmax2
            val     = x(k) - myo.params(10);
            isterm  = 1;
            Dir     = 1;
        end
        
        function [val, isterm, Dir] = noevent(obj,t,x,myo,k)
            % Represents no event. Dimension 2 to maintain a uniform
            % dimension to event functions between modes and avoid annoying
            % computations in the heart's jump function
            val     = [1, 1];
            isterm  = [0, 0];
            Dir     = [0, 0];
        end
        
        function ng = nb_guards(obj,mode)
            ng = length(obj.transitions(mode));
        end
        
        function [newmode, xr, myo] = jump(obj, activatedGuard, activationTime, myo, x, k)
            % [newmode, xr] = jump(obj, activatedGuard, activationTime, myo, x, k)
            % Transition to new location from current mode based on the
            % index in myguard. myguard is (part of) the IE vector returned
            % by ode45 (and other solvers).
            % myo is the myocyte being updated.
            % x is the state that caused the event (i.e. that satisfied the
            % guard): x [voltage, timer, dV/dt]
            % k is the linear index of myocyte
            % OUTPUTS:
            % newmode is the new mode
            % xr is post-reset state
            % myo is the input myocyte with its parameters updated
            global stat;
            % Only one event should be active for determinism
            assert(length(activatedGuard) == 1)
            mode = myo.mode;
            source = obj.transitions(mode);
            newmode = source(activatedGuard);
            myo.mode = newmode;
            % Resets
            xr = x;
            if newmode == obj.UPSTROKE
                xr(3) = myo.params(obj.ixUPSTROKE_SLOPE);
            elseif newmode == obj.UPSTROKE2
                xr(3) = myo.params(obj.ixUPSTROKE2_SLOPE);
            elseif newmode == obj.NOTCH
                xr(3) = myo.params(obj.ixNOTCH_SLOPE);            
            elseif newmode == obj.PLATEAU
                % reset timer to enable measuring time elapsed in plateau mode
                xr(2) = 0;
                xr(3) = 0;
            elseif newmode == obj.ERP
                xr(3) = myo.params(obj.ixERP_SLOPE);
            elseif newmode == obj.UPSTROKE2
                xr(3) = myo.params(obj.ixUPSTROKE2_SLOPE);    
            end
            
            % Update variables needed for restitution
            if obj.use_restitution && (newmode == obj.UPSTROKE || newmode == obj.UPSTROKE2)               
                % Calculate the previous DI and so the new APD
                % - Record upstroke time for upstroking cells
                myo.upstroke_time = activationTime; 
                % - Record end of previous AP
                if myo.APDwasset
                    myo.DI = myo.upstroke_time - myo.endofAPD;
                elseif myo.troughWasset
                     % APD might not be set if APD90 is never reached in
                     % previous AP, which causes issues when computing DI
%                     warning('t=%d, APD was not set!', activationTime)
                    myo.DI = myo.upstroke_time - myo.ttrough;                    
                else
                    % nothing changes
%                     warning('t = %d, Trough and APD not set! DI still = %d\n', activationTime, myo.DI)
                end                               
                newAPD = max(obj.restitution(myo.DI), myo.params(obj.ixAPDMIN)) ;                
                % Rather than make d = upstroke slope a function of Vtrough (which
                % is the smallest voltage reached before depolarization),
                % we use a shortcut and make it a function of Vth because
                % accessing Vtrough here is a b*
                dmax    = myo.params(obj.ixUPSTROKE_SLOPE_MAX);
                dmin    = myo.params(obj.ixUPSTROKE_SLOPE_MIN);
                Vmin    = myo.params(obj.ixVMIN);
                switch newmode
                    case obj.UPSTROKE                        
                        Vth     = myo.params(obj.ixVTH);
                        d       = dmax + ((dmin-dmax)/(Vth - Vmin))*(x(1) - Vmin);
                        myo.params(obj.ixUPSTROKE_SLOPE) = d;                        
                    case obj.UPSTROKE2
                        Vth     = myo.params(obj.ixVTH2);
                        d       = dmax + ((dmin-dmax)/(Vth - Vmin))*(x(1) - Vmin);
                        myo.params(obj.ixUPSTROKE2_SLOPE) = d;
                end                
                % newAPD = upstroke_duration + Plateau duration + downstroke_duration
                upstroke_duration   = (myo.params(obj.ixVMAX) - Vth) /d;
                downstroke_duration = (Vmin - myo.params(obj.ixVPLATEAU)) / myo.params(obj.ixERP_SLOPE);
                newPlD = newAPD - upstroke_duration - downstroke_duration;
                assert(newPlD > 0, 'New plateau duration is negative!');
                myo.params(obj.ixPLATEAU_DURATION) = newPlD;
                stat{k} = [stat{k};myo.APDwasset, myo.endofAPD, myo.troughWasset, myo.ttrough, ...
                    myo.params(obj.ixPLATEAU_DURATION), myo.DI, newAPD, d];
                % reset to track whether it gets set at end of this AP
                myo.APDwasset = 0; 
                myo.troughWasset = 0;
            end
            
        end
        
        %% Gradients
        function dx = gradientQUIESCENT(obj,t,x,myo,k)
            nh = length(myo.horiz_nhbrs); % nh=2
            nv = length(myo.verti_nhbrs); %n=2
            dx(1) = (sum(x(myo.horiz_nhbrs))-nh*x(k))/myo.Rh + (sum(x(myo.verti_nhbrs))-nv*x(k))/myo.Rv;
            dx(2) = 1;
            n2 = 2*obj.n^2;
            dx(3) = (sum(x(myo.horiz_nhbrs+n2))-nh*x(k+n2))/myo.Rh + ...
                (sum(x(myo.verti_nhbrs+n2))-nv*x(k+n2))/myo.Rv;
        end
        
        function dx = gradient_selfexcQUIESCENT(obj,t,x,myo,k)
            % A self-exciting cell's voltage gradient is modeled as having
            % an intrinsic positive value
            nh = length(myo.horiz_nhbrs); % nh=2
            nv = length(myo.verti_nhbrs); %n=2
            dx(1) = obj.Vinq + (sum(x(myo.horiz_nhbrs))-nh*x(k))/myo.Rh + (sum(x(myo.verti_nhbrs))-nv*x(k))/myo.Rv;
            dx(2) = 1;
            n2 = 2*obj.n^2;
            dx(3) = (sum(x(myo.horiz_nhbrs+n2))-nh*x(k+n2))/myo.Rh + ...
                (sum(x(myo.verti_nhbrs+n2))-nv*x(k+n2))/myo.Rv;
        end
        
        function dx = gradientUPSTROKE(obj,t,x,myo,k)
            dx(1) = myo.params(obj.ixUPSTROKE_SLOPE);
            dx(2) = 1;
            dx(3) = 0;
        end
        
        function dx = gradient_selfexcUPSTROKE(obj,t,x,myo,k)
            dx(1) = myo.params(obj.ixUPSTROKE_SLOPE);
            dx(2) = 1;
            dx(3) = 0;
        end
        
        function dx = gradientNOTCH(obj,t,x,myo,k)
            dx(1) = myo.params(obj.ixNOTCH_SLOPE);
            dx(2) = 1;
            dx(3) = 0;
        end
        
        function dx = gradientPLATEAU(obj,t,x,myo,k)
            dx(1) = 0;
            dx(2) = 1;
            dx(3) = 0;
        end
        
        function dx = gradientERP(obj,t,x,myo,k)
            dx(1) = myo.params(obj.ixERP_SLOPE);
            dx(2) = 1;
            dx(3) = 0;
        end
        
        function dx = gradientRRP(obj,t,x,myo,k)
            nh = length(myo.horiz_nhbrs); % nh=2
            nv = length(myo.verti_nhbrs); %n=2
            dx(1) = obj.Vinrrp + (sum(x(myo.horiz_nhbrs))-nh*x(k))/myo.Rh + (sum(x(myo.verti_nhbrs))-nv*x(k))/myo.Rv;
            dx(2) = 1;
            n2 = 2*obj.n^2;
            dx(3) = (sum(x(myo.horiz_nhbrs+n2))-nh*x(k+n2))/myo.Rh + ...
                (sum(x(myo.verti_nhbrs+n2))-nv*x(k+n2))/myo.Rv;
        end
        
        function dx = gradientUPSTROKE2(obj,t,x,myo,k)
            dx(1) = myo.params(obj.ixUPSTROKE2_SLOPE);
            dx(2) = 1;
            dx(3) = 0;
        end
        
        %% Nhbrd structure
        function [horiz_cell_nhbrs, verti_cell_nhbrs] = create_4nhbr(obj, n)
            % Structure of neighbors that contribute to cell
            % cell_nhbrs(i,j) is an array, giving the LINEAR indices
            % of the neighbors.
            % The following code gives a 4-nhbd
            verti_cell_nhbrs = cell(n);
            horiz_cell_nhbrs = cell(n);
            SIZ = [n,n];
            % First the border cases
            % 1st row
            horiz_cell_nhbrs{1,1} = sub2ind(SIZ,1,2);
            verti_cell_nhbrs{1,1} = sub2ind(SIZ,2,1);
            horiz_cell_nhbrs{1,n} = sub2ind(SIZ,1,n-1);
            verti_cell_nhbrs{1,n} = sub2ind(SIZ,2,n);
            for j=2:n-1
                horiz_cell_nhbrs{1,j} = sub2ind(SIZ,[1, 1], [j-1, j+1]);
                verti_cell_nhbrs{1,j} = sub2ind(SIZ,2, j);
            end
            % last row
            horiz_cell_nhbrs{n,1} = sub2ind(SIZ,n,2);
            verti_cell_nhbrs{n,1} = sub2ind(SIZ,n-1,1);
            horiz_cell_nhbrs{n,n} = sub2ind(SIZ,n, n-1);
            verti_cell_nhbrs{n,n} = sub2ind(SIZ,n-1,n);
            for j=2:n-1
                horiz_cell_nhbrs{n,j} = sub2ind(SIZ,[n, n], [j-1, j+1]);
                verti_cell_nhbrs{n,j} = sub2ind(SIZ,n-1, j);
            end
            % first column (minus 1st and last row)
            for i=2:n-1
                horiz_cell_nhbrs{i,1} = sub2ind(SIZ,i, 2);
                verti_cell_nhbrs{i,1} = sub2ind(SIZ, [i-1, i+1], [1, 1]);
            end
            % last column (minus 1st and last row)
            for i=2:n-1
                horiz_cell_nhbrs{i,n} = sub2ind(SIZ,i, n-1);
                verti_cell_nhbrs{i,n} = sub2ind(SIZ, [i-1, i+1], [n, n]);
            end
            % middle of grid
            for i=2:n-1
                for j=2:n-1
                    horiz_cell_nhbrs{i,j} = sub2ind(SIZ,[i-1, i+1], [j, j]);
                    verti_cell_nhbrs{i,j} = sub2ind(SIZ, [i, i], [j-1, j+1]);
                end
            end
        end
        
        %% Random params
        % Create random arrays of parameters
        function params = generate_params(obj, n, randomized)
            % params = generate_params(obj, n)
            % params = generate_params(obj, n, randomized[0|1])
            if nargin < 3
                randomized = 1;
            end
            params = zeros(n,n,16);
            % d (current upstroke slope, phase 0)
            params(:,:,obj.ixUPSTROKE_SLOPE) = 6 + randomized*0.1*(-1 + 2*rand(n));
            % g (notch, phase 1)
            params(:,:,obj.ixNOTCH_SLOPE) = -1 + randomized*0.05*(-1 + 2*rand(n));
            % b (phase 3 ERP)
            params(:,:,obj.ixERP_SLOPE) = -2 + randomized*0.1*(-1 + 2*rand(n));
            % d2 (upstroke 2)
            params(:,:,obj.ixUPSTROKE2_SLOPE) = 3 + randomized*0.1*(-1 + 2*rand(n));
            % Vth for depolarization
            params(:,:,obj.ixVTH) = -80 + randomized*2*(-1 + 2*rand(n));
            % Vmax
            params(:,:,obj.ixVMAX) = 52 + randomized*2*(-1 + 2*rand(n));
            % Vplateau (transition from Notch to Plateau)
            params(:,:,obj.ixVPLATEAU) = 0.9*params(:,:,obj.ixVMAX);
            % PD (plateau duration)
            params(:,:,obj.ixPLATEAU_DURATION) = 100 + randomized*10*(-1 + 2*rand(n));
            % Vth2 > Vth (remember Vth is negative)
            params(:,:,obj.ixVTH2) = 0.7*params(:,:,obj.ixVTH);
            % Vmax2 < Vmax
            params(:,:,obj.ixVMAX2) = 0.7*params(:,:,obj.ixVMAX);
            % Vmin (voltage when quiescent)
            params(:,:,obj.ixVMIN) = -92 + randomized*2*(-1 + 2*rand(n));
            % Vrrp (transition from ERP to RRP when V(i,j) <= Vrrp)
            params(:,:,obj.ixVRRP) = 0.8*params(:,:,obj.ixVTH);
            % APDmax
            params(:,:,obj.ixAPDMAX) = 600 + randomized*10*(-1 + 2*rand(n));
            % APDmin
            params(:,:,obj.ixAPDMIN) = 0.5*params(:,:,obj.ixAPDMAX);
            % dmin = min upstroke slope
            params(:,:,obj.ixUPSTROKE_SLOPE_MIN) = 0.5*params(:,:,obj.ixUPSTROKE_SLOPE);
            % dmax = max upstroke velocity. When initializes, starts at max
            % upstroke slope
            params(:,:,obj.ixUPSTROKE_SLOPE_MAX) = params(:,:,obj.ixUPSTROKE_SLOPE);
        end
        
        function b = trough_detection_mode(obj,mode)
            b= mode == obj.QUIESCENT || mode == obj.RRP;
        end
    end
end